// https://cantwell-tom.medium.com/ellers-maze-algorithm-in-javascript-2e5742c1a4cd
// Eller's algorithm
// 1) Initialize the first row
// 2) For each cell, add it to a new set unless the cell above it is open downwards
// 3) For each cell, randomly remove right and/or bottom walls, merging sets
// 4) ensure that each set has at least one open wall down
// 5) create a new row and go to step 2
// 6) the last row is different, always open a wall between two sets

function shouldRemoveWall(x, y) {
    // this seems to generate interesting mazes
    let factor = 0.5
    if ((y % 2) == 0) {
        factor = 0.60
    }
    if ((x % 2) == 0) {
        factor = 0.75
    }
    return Math.random() < factor
}

export function ellerMaze(sizeX, sizeY) {
    let maze = []

    let sets = new Map()

    for (let y = 0; y < sizeY; ++y) {
        let row = []
        maze[y] = row

        for (let x = 0; x < sizeX; ++x) {
            let setName = `${x}:${y}`
            let cell = { x: x, y: y, set: setName, connections: {} }
            if ((x > 0) && (row[x - 1].connections.right !== undefined)) {
                setName = row[x - 1].set
                cell.connections.left = true
                cell.set = setName
                // console.log("setName: ", setName, sets.get(setName))
                sets.get(setName).push(cell)
            } else {
                sets.set(setName, [cell])
            }

            if ((y > 0) && (maze[y - 1][x].connections.down !== undefined)) {
                // merge sets from above
                cell.connections.up = true
                let oldSetName = cell.set
                let newSetName = maze[y - 1][x].set
                if (oldSetName != newSetName) {
                    // console.log(`merging ${oldSetName} to ${newSetName}`)
                    sets.get(oldSetName).forEach(c => {
                        c.set = newSetName
                        sets.get(newSetName).push(c)
                    });
                    sets.delete(oldSetName)
                }
            }

            if (x < sizeX - 1) {
                if (shouldRemoveWall()) {
                    cell.connections.right = true
                }
            }
            row[x] = cell
        }

        // make a temporary map of all the sets in the current row, this has to be done
        // after all the merging has been done
        let rowSets = new Map()
        row.forEach(c => {
            if (rowSets.has(c.set)) {
                rowSets.get(c.set).push(c)
            } else {
                rowSets.set(c.set, [c])
            }
        })

        // ensure each rowSet has at least one exit down
        if (y < sizeY - 1) {
            rowSets.forEach(cells => {
                if (cells.length == 1) {
                    cells[0].connections.down = true
                } else {
                    let removed = false
                    while (removed == false) {
                        cells.forEach(c => {
                            if (shouldRemoveWall()) {
                                c.connections.down = true
                                removed = true
                            }
                        })
                    }
                }
            })
        } else {
            // last row, break walls between different sets
            for (let x = 1; x < sizeX; ++x) {
                if (row[x].set != row[x - 1].set) {
                    row[x].connections.left = true
                    row[x - 1].connections.right = true
                }
            }
        }
    }
    return maze
}
