import { StarChart } from "./starchart.js"
import { ellerMaze } from "./ellermaze.js"
import * as Cards from "./cards.js"

let mostRecentDialogDiv = null
let starChartDisplay = null

class Sector {
    constructor(x, y, connections) {
        this.x = x
        this.y = y
        this.connections = connections
        this.visited = false
        this.state = {}
        this.layoutRevealed = false
        this.knot = "placeholder"
        this.distanceToEnd = -1
    }

    numberOfExits() {
        let count = 0
        if (this.connections.up === true) count++;
        if (this.connections.down === true) count++;
        if (this.connections.left === true) count++;
        if (this.connections.right === true) count++;

        return count;
    }
}

// represents a rectangular range of sectors
class SectorRange {
    constructor(minX, minY, maxX, maxY) {
        this.min = { x: minX, y: minY }
        this.max = { x: maxX, y: maxY }
    }

    inRange(x, y) {
        return (x >= this.min.x) && (x <= this.max.x) && (y >= this.min.y) && (y <= this.max.y);
    }
}

// represents an explicit list of sectors
class OddSectorRange {
    constructor(possibleSectors) {
        this.possibleSectors = possibleSectors
    }

    inRange(x, y) {
        return (this.possibleSectors.find(e => (e.x == x && e.y == y)) != undefined)
    }
}

function shuffleArray(arr) {
    for (let i = arr.length - 1; i > 0; i--) {
        const j = Math.floor(Math.random() * (i + 1));
        [arr[i], arr[j]] = [arr[j], arr[i]]
    }
}

class GameState {
    constructor(story, sizeX, sizeY) {
        this.story = story
        this.dimensions = { x: sizeX, y: sizeY }
        this.shipPath = []

        // try to make interesting mazes by rejecting those that can be solved
        // in only a few moves
        while (true) {
            let maze = ellerMaze(sizeX, sizeY)

            this.map = new Array(this.dimensions.x)
            for (let i = 0; i < this.dimensions.y; ++i) {
                this.map[i] = Array(this.dimensions.y)
                for (let j = 0; j < this.map[i].length; ++j) {
                    this.map[i][j] = new Sector(i, j, maze[j][i].connections)
                }
            }

            // the first sector should be open north and east
            this.removeMazeWall("A4", 1) // north
            this.removeMazeWall("A4", 2) // east

            // find the distances to the last sector
            let distances = this.findMazeDistance({ x: sizeX - 1, y: 0 })
            for (let i = 0; i < this.dimensions.y; ++i) {
                for (let j = 0; j < this.map[i].length; ++j) {
                    let t = this.getSectorXY({ x: i, y: j })
                    t.distanceToEnd = distances[i][j]
                }
            }

            let shortestDistance = distances[0][this.dimensions.y - 1]
            if (shortestDistance >= 11 && shortestDistance <= 18) {
                break;
            }
        }

        const allSectors = new SectorRange(0, 0, 4, 4)
        const middleSectors = new SectorRange(1, 1, 3, 3)
        const towardsTheEnd = new SectorRange(2, 0, 4, 2)
        const towardsTheStart = new SectorRange(0, 2, 2, 4)
        const topLeft = new SectorRange(0, 0, 1, 1)
        const bottomRight = new SectorRange(3, 3, 4, 4)
        const middleSectorsExceptD1 = new OddSectorRange([{ x: 1, y: 1 }, { x: 2, y: 1 },
        { x: 1, y: 2 }, { x: 2, y: 2 }, { x: 3, y: 2 },
        { x: 1, y: 3 }, { x: 2, y: 3 }, { x: 3, y: 3 }])

        const allKnots = [
            { knot: "asteroid_lair", range: towardsTheStart, goAnywhere: true },
            { knot: "lost_probe", range: topLeft, goAnywhere: true },
            { knot: "harvest_fuel_from_gas_giant", range: towardsTheEnd, goAnywhere: true },
            { knot: "wormhole", range: "special", goAnywhere: true },
            { knot: "eden", range: bottomRight, goAnywhere: true },
            { knot: "stasis_pod", range: towardsTheStart, goAnywhere: true },
            { knot: "bridge_invaded", range: bottomRight, goAnywhere: true },
            { knot: "space_storm", range: allSectors, goAnywhere: true },
            { knot: "space_station", range: middleSectors, goAnywhere: true },
            { knot: "asteroid_belt", range: topLeft, goAnywhere: true },
            { knot: "pressure_wave", range: topLeft, goAnywhere: false },
            { knot: "trader", range: towardsTheEnd, goAnywhere: true },
            { knot: "neutron_star", range: middleSectorsExceptD1, goAnywhere: false },
            { knot: "derelict_ship", range: towardsTheEnd, goAnywhere: true },
            { knot: "missing_time", range: middleSectors, goAnywhere: true },
            { knot: "patrol_lane", range: towardsTheEnd, goAnywhere: true },
            { knot: "antibodies", range: middleSectors, goAnywhere: true },
            { knot: "plague_planet", range: towardsTheEnd, goAnywhere: true },
            { knot: "statue", range: towardsTheStart, goAnywhere: true },
            { knot: "gate", range: bottomRight, goAnywhere: false },
            { knot: "thin_walls", range: middleSectors, goAnywhere: false },
            { knot: "primitive_planet", range: towardsTheStart, goAnywhere: true },
            { knot: "glexx_fight1", range: towardsTheStart, goAnywhere: true },
            { knot: "glexx_dreadnought", range: towardsTheEnd, goAnywhere: true },
            { knot: "empty_sector2", range: allSectors, goAnywhere: true }
        ]

        let availableKnots = new Set()
        allKnots.forEach(t => {
            availableKnots.add(t.knot)
        })

        if ((Math.random() * 10) > 3) {
            // the pressure wave should go where it would be most useful
            // find the sector farthest from the end with a blocked way east
            let biggestDifferentEast = 0
            let furthestSector = this.map[4][0]
            for (let y = 0; y < this.dimensions.y; ++y) {
                for (let x = 0; x < this.dimensions.x - 1; ++x) {  // cannot go in right column
                    if ((x <= 1) && (y >= 3)) {
                        continue; // first sectors no good
                    }
                    let s = this.map[x][y]
                    let sectorToEast = this.map[x + 1][y]
                    if (sectorToEast.distanceToEnd <= 3) {
                        continue;   // pressure wave doesn't open up too close to the exit
                    }
                    let diff = s.distanceToEnd - sectorToEast.distanceToEnd
                    if (diff > biggestDifferentEast) {
                        biggestDifferentEast = diff
                        furthestSector = s
                    }
                }
            }
            if (furthestSector != this.map[4][0]) {
                furthestSector.knot = "pressure_wave"
                availableKnots.delete("pressure_wave")
            }
        }

        {
            // place the gate somewhere interesting
            let biggestDifference = -1
            let bestSectors = []
            for (let y = 1; y < this.dimensions.y; ++y) { // gate cannot go in top row
                for (let x = 0; x < this.dimensions.x; ++x) {
                    if ((x <= 1) && (y >= 3)) {
                        continue // first sectors no good
                    }
                    let s = this.map[x][y]
                    if (s.connections.up === true) {
                        continue    // need a wall to the north
                    }

                    let distances = this.findMazeDistance({ x: x, y: y - 1 })
                    if (distances[x][y] == biggestDifference) {
                        console.log("Best distance so far ", x, ", ", y, " = ", biggestDifference)
                        bestSectors.push(s)
                    } else if (distances[x][y] > biggestDifference) {
                        console.log("Best distance so far ", x, ", ", y, " = ", biggestDifference)
                        bestSectors = [s]
                        biggestDifference = distances[x][y]
                    }
                }
            }
            if (bestSectors.length > 0) {
                shuffleArray(bestSectors)
                console.log(bestSectors)
                bestSectors[0].knot = "gate"
                availableKnots.delete("gate")
            }
        }

        // some special cases
        this.map[4][0].knot = "last_sector"
        this.map[0][4].knot = "first_sector"
        //this.map[0][4].knot = "empty_sector2"

        let scrambledSectors = this.getUnassignedSectors()

        scrambledSectors.forEach(sector => {
            let mapSector = this.map[sector.x][sector.y]

            if (mapSector.knot !== "placeholder") {
                return; // already assigned
            }

            // filter possible knots
            let possibleKnots = allKnots.filter(g => {
                if (availableKnots.has(g.knot) && (g.range instanceof SectorRange)) {
                    return g.range.inRange(sector.x, sector.y)
                } else {
                    return false;
                }
            })

            // gate needs to be placed somewhere without a passage north
            if (mapSector.connections.up === true) {
                possibleKnots = possibleKnots.filter(g => {
                    return g.knot !== "gate"
                })
            }

            // thin walls needs somewhere with walls
            if (mapSector.numberOfExits() >= 3) {
                possibleKnots = possibleKnots.filter(g => {
                    return g.knot !== "thin_walls"
                })
            }

            // show preference for thin walls where the way is blocked north or east
            if (mapSector.numberOfExits() <= 2) {
                if (possibleKnots.includes("thin_walls") &&
                    ((mapSector.connections.up !== true) && (mapSector.connections.right !== true))) {
                    console.log("selected thin_walls in", mapSector.x, mapSector.y)
                    mapSector.knot = "thin_walls"
                    availableKnots.delete("thin_walls")
                    return
                }
            }

            // show preference to the wormhole for far away sectors
            if ((mapSector.distanceToEnd > 14) && (availableKnots.has("wormhole") && (mapSector.x >= 1 && mapSector.y <= 3))) {
                console.log("selected wormhole")
                mapSector.knot = "wormhole"
                availableKnots.delete("wormhole")
                return
            }

            // show preference to statue for sectors with lots of exits
            if ((mapSector.numberOfExits() >= 3) && (availableKnots.has("statue"))) {
                console.log("selected statue in", mapSector.x, mapSector.y)
                mapSector.knot = "statue"
                availableKnots.delete("statue")
                console.log(this.getBestExitFromSector(this.convertXYToSector({ x: mapSector.x, y: mapSector.y })))
                return
            }

            // missing time should prioritize dead ends
            if ((mapSector.numberOfExits() == 1) && (availableKnots.has("missing_time"))) {
                console.log("missing time selected")
                mapSector.knot = "missing_time"
                availableKnots.delete("missing_time")
                return
            }

            if (possibleKnots.length == 0) {
                console.log("No possible knot for ", mapSector.x, mapSector.y)
                return; // no possible knots for this sector
            }

            let selectedKnot = possibleKnots[Math.floor(Math.random() * possibleKnots.length)]
            mapSector.knot = selectedKnot.knot
            availableKnots.delete(selectedKnot.knot)
        })


        scrambledSectors = this.getUnassignedSectors()
        let possibleKnots = allKnots.filter(g => (availableKnots.has(g.knot) && (g.goAnywhere == true)))
        for (let i = 0; i < Math.min(scrambledSectors.length, possibleKnots.length); ++i) {
            let mapSector = this.map[scrambledSectors[i].x][scrambledSectors[i].y]
            mapSector.knot = possibleKnots[i].knot
            console.log("Placing unassigned knot ", mapSector.knot, " in ", mapSector)
        }
    }

    getUnassignedSectors() {
        // create a scrambled array of valid sectors
        let sectors = new Array()
        for (let y = 0; y < this.dimensions.y; ++y) {
            for (let x = 0; x < this.dimensions.x; ++x) {
                if (this.map[x][y].knot == "placeholder") {
                    sectors.push({ x: x, y: y });
                }
            }
        }
        shuffleArray(sectors)
        return sectors;
    }

    getShipPath() {
        return this.shipPath
    }

    getDeltaForDirection(direction) {
        let delta = { x: 0, y: 0 }
        switch (direction) {
            case 1:
                delta.y -= 1;
                break;
            case 2:
                delta.x += 1;
                break;
            case 3:
                delta.y += 1;
                break;
            case 4:
                delta.x -= 1;
                break;
        }
        return delta
    }

    isValidSector(sectorXY) {
        return (sectorXY.x >= 0 && sectorXY.y < this.dimensions.x && sectorXY.y >= 0 && sectorXY.y < this.dimensions.y)
    }

    isWayBlocked(shipSector, direction) {
        let t = this.getSectorXY(this.convertSectorToXY(shipSector))
        switch (direction) {
            case 1:
                return (t.connections.up !== true)
            case 2:
                return (t.connections.right !== true)
            case 3:
                return (t.connections.down !== true)
            case 4:
                return (t.connections.left !== true)
        }
    }

    findMazeDistance(targetSectorXY) {
        let shortestDistanceSoFar = new Array(this.dimensions.x)
        let sectorQueue = [targetSectorXY]
        for (let i = 0; i < shortestDistanceSoFar.length; ++i) {
            shortestDistanceSoFar[i] = new Array(this.dimensions.y)
            shortestDistanceSoFar[i].fill(10000)
        }
        shortestDistanceSoFar[targetSectorXY.x][targetSectorXY.y] = 0

        let evalSector = (s, currentCost) => {
            let c = shortestDistanceSoFar[s.x][s.y]
            if (c > currentCost) {
                shortestDistanceSoFar[s.x][s.y] = currentCost + 1
                sectorQueue.push(s)
            }
        }

        while (sectorQueue.length != 0) {
            let sectorXY = sectorQueue.shift()
            let sector = this.getSectorXY(sectorXY)
            let cost = shortestDistanceSoFar[sector.x][sector.y]
            if (sector.connections.up === true) {
                let t = this.getSectorXY({ x: sectorXY.x, y: sectorXY.y - 1 })
                evalSector(t, cost)
            }
            if (sector.connections.down === true) {
                let t = this.getSectorXY({ x: sectorXY.x, y: sectorXY.y + 1 })
                evalSector(t, cost)
            }
            if (sector.connections.left === true) {
                let t = this.getSectorXY({ x: sectorXY.x - 1, y: sectorXY.y })
                evalSector(t, cost)
            }
            if (sector.connections.right === true) {
                let t = this.getSectorXY({ x: sectorXY.x + 1, y: sectorXY.y })
                evalSector(t, cost)
            }
        }
        return shortestDistanceSoFar
    }

    removeMazeWall(shipSector, direction) {
        let t = this.getSectorXY(this.convertSectorToXY(shipSector))
        let delta = this.getDeltaForDirection(direction)
        let newSectorXY = { x: t.x + delta.x, y: t.y + delta.y }
        if (!this.isValidSector(newSectorXY)) {
            return false;
        }

        let newSector = this.getSectorXY(newSectorXY)

        switch (direction) {
            case 1:
                t.connections.up = true
                newSector.connections.down = true
                break
            case 2:
                t.connections.right = true
                newSector.connections.left = true
                break
            case 3:
                t.connections.down = true
                newSector.connections.up = true
                break
            case 4:
                t.connections.left = true
                newSector.connections.right = true
                break
        }
        if (starChartDisplay != undefined) {
            starChartDisplay.draw()
        }

        return true
    }

    getSectorsSurrounding(sectorXY) {
        let result = []
        for (let y = Math.max(sectorXY.y - 1, 0); y <= (Math.min(sectorXY.y + 1, this.dimensions.y - 1)); ++y) {
            for (let x = Math.max(sectorXY.x - 1, 0); x <= (Math.min(sectorXY.x + 1, this.dimensions.x - 1)); ++x) {
                result.push({ x: x, y: y })
            }
        }
        return result
    }

    revealSectorsSurroundingShip() {
        let sectors = this.getSectorsSurrounding(this.getCurrentSectorXY())
        let count = 0
        sectors.forEach((s => {
            let t = this.getSectorXY(s)
            if (t.layoutRevealed == false) {
                count++;
            }
            t.layoutRevealed = true
        }).bind(this))
        if (starChartDisplay != undefined) {
            starChartDisplay.draw()
        }

        return count
    }

    revealSector(coord) {
        let t = this.getSectorXY(this.convertSectorToXY(coord))
        let old = t.layoutRevealed
        t.layoutRevealed = true;
        if (starChartDisplay != undefined) {
            starChartDisplay.draw()
        }
        return (old == false)
    }

    getSectorAtOffset(deltaX, deltaY) {
        let s = this.getCurrentSectorXY()
        let x = s.x + deltaX
        let y = s.y + deltaY
        if ((x < 0) || (x >= this.dimensions.x) || (y < 0) || (y >= this.dimensions.y)) {
            return "";
        }
        return this.convertXYToSector({ x: x, y: y })
    }

    findRandomAdjacentSector(shipSector) {
        let validDestinations = []
        for (let i = 1; i <= 4; ++i) {
            let newSector = this.canMove(shipSector, i)
            if (newSector !== "") {
                validDestinations.push(newSector)
                // favor unvisited destinations by pushing twice
                if (!this.isVisited(newSector)) {
                    validDestinations.push(newSector)
                }
            }
        }
        return validDestinations[Math.floor(Math.random() * validDestinations.length)];
    }

    findSuitableWormholeExit(shipSector, level) {
        let t = this.getSectorXY(this.convertSectorToXY(shipSector))
        let distancesFromHere = this.findMazeDistance(t)

        // try and put the exit a long way away from here, in a place we haven't visited
        // and not too close to the exit
        for (let attempt = 0; attempt < 20; ++attempt) {
            let x = Math.floor(Math.random() * this.dimensions.x)
            let y = Math.floor(Math.random() * this.dimensions.y)
            let sector = this.map[x][y]
            if (sector.distanceToEnd <= 5) {
                continue  // too close to end
            }
            if ((sector.x % 2) != level) {
                continue; // make sure that the different levels have different results
            }
            if (level == 0) {
                if (sector.distanceToEnd > t.distanceToEnd) {
                    continue // don't want to send player further away
                }
            }
            if (sector.visited == true) {
                continue // don't want to send player to a place they have been
            }
            if (distancesFromHere[x][y] < 3) {
                continue // want a sector at least 3 sectors away from here
            }

            return this.convertXYToSector({ x: sector.x, y: sector.y })
        }
        // screw it, into the corner you go
        if (level == 0) {
            return this.convertXYToSector({ x: 0, y: 0 })
        } else {
            return this.convertXYToSector({ x: 4, y: 4 })
        }
    }

    getDimensions() {
        return this.dimensions
    }

    getSectorAt(x, y) {
        return this.map[x][y]
    }

    getSectorXY(xy) {
        return this.getSectorAt(xy.x, xy.y)
    }

    convertSectorToXY(coords) {
        let c = coords.toLowerCase()
        let x = c.charCodeAt(0) - "a".charCodeAt(0)
        let y = c.charCodeAt(1) - "0".charCodeAt(0)
        return { "x": x, "y": y }
    }

    convertXYToSector(xy) {
        let s = String.fromCharCode(xy.x + "A".charCodeAt(0))
        s += xy.y.toString()
        return s
    }

    getCurrentSectorXY() {
        let storySector = this.story.variablesState["sector"].toLowerCase()
        let xy = this.convertSectorToXY(storySector)
        return xy
    }

    canMove(fromSector, direction) {
        let s = this.getSectorXY(this.convertSectorToXY(fromSector))

        let newXY = { x: s.x, y: s.y }
        if (direction == 1 && s.connections.up != undefined) {
            newXY.y -= 1
            return this.convertXYToSector(newXY)
        } else if (direction == 2 && s.connections.right != undefined) {
            newXY.x += 1
            return this.convertXYToSector(newXY)
        } else if (direction == 3 && s.connections.down != undefined) {
            newXY.y += 1
            return this.convertXYToSector(newXY)
        } else if (direction == 4 && s.connections.left != undefined) {
            newXY.x -= 1
            return this.convertXYToSector(newXY)
        } else {
            console.log("Unexpected direction ", direction)
        }
        return ""
    }

    isVisited(sector) {
        let s = this.getSectorXY(this.convertSectorToXY(sector))

        return s.visited
    }

    jumpedToPacifica() {
        this.shipPath.push({ x: -1, y: -1, style: "pacifica" });
        if (starChartDisplay != undefined) {
            starChartDisplay.draw()
        }
    }

    jumpToNewSector() {
        let specialMovement = this.story.variablesState["specialMovement"]
        this.story.variablesState["specialMovement"] = false
        let newSector = this.getSectorXY(this.getCurrentSectorXY())
        this.shipPath.push({ x: newSector.x, y: newSector.y, style: specialMovement ? "special" : "normal" });

        if (!newSector.visited) {
            newSector.visited = true
            newSector.layoutRevealed = true
        }
        if (starChartDisplay != undefined) {
            starChartDisplay.draw()
        }
        this.story.ChoosePathString(newSector.knot)
    }

    formatStardate(s) {
        return s.toFixed(2)
    }

    getBestExitFromSector(shipSector) {
        let surroundingSectors = []
        for (let i = 1; i <= 4; ++i) {
            let adjacent = this.canMove(shipSector, i)
            if (adjacent !== "") {
                surroundingSectors.push(
                    this.getSectorXY(this.convertSectorToXY(adjacent)))
            }
        }
        surroundingSectors.sort((a, b) => {
            let d = a.distanceToEnd - b.distanceToEnd
            if (d != 0) {
                return d
            }
            return a.knot < b.knot
        })

        return this.convertXYToSector({ x: surroundingSectors[0].x, y: surroundingSectors[0].y });
    }
}
// exported for debugging

export let gameState = null

// Creates a dialog paragraph, usually of the form
// Character: some words they spoke
function createDialogElement(text) {
    let re = /^(.+?:)(.*)/
    let matches = text.match(re)

    console.log(text, matches)

    let element = document.createElement("p")
    element.classList.add("dialog")
    if ((matches == null) || (matches.length !== 3)) {
        element.innerHTML = text
    } else {
        let character = document.createElement("span")
        character.classList.add("char")
        character.innerText = matches[1]
        let words = document.createElement("span")
        words.classList.add("words")
        words.innerHTML = matches[2]
        element.appendChild(character)
        element.appendChild(words)
    }
    return element
}

function setViewportForButton() {
    const buttons = document.getElementsByName("displaymode")

    for (let i = 0; i < buttons.length; ++i) {
        let target = document.getElementById(buttons[i].value)
        if (buttons[i].checked) {
            if (target.classList.contains("notselected")) {
                target.classList.remove("notselected")
                target.classList.add("selected")
                window.scroll(0, (target.savedScrollPos == null) ? 0 : target.savedScrollPos)
            }
        } else {
            if (target.classList.contains("selected")) {
                target.savedScrollPos = window.scrollY
                target.classList.remove("selected")
                target.classList.add("notselected")
            }
        }
    }
    return true
}

function removeChoicesFromPage(parentElement) {
    // explicitly set the height of the content element so that removing elements doesn't cause the page to get shorter as possibly
    // scroll in the distracting way
    const bottomElement = parentElement.lastElementChild;
    const height = bottomElement ? bottomElement.offsetTop + bottomElement.offsetHeight : 0;
    parentElement.style.height = height + "px"

    const choicesList = Array.from(parentElement.getElementsByClassName("choicelist"))
    choicesList.forEach((e) => {
        e.parentNode.removeChild(e)
    })
}

function removeEverythingFromPage(parentElement) {
    while (parentElement.firstChild) {
        parentElement.removeChild(parentElement.lastChild)
    }
    parentElement.style.height = "auto"
}

function insertImage(parentElement, imageName, description, cssClass) {
    //<as-dithered-image src="mypicture.jpg" alt="Description of the image for screen readers"></as-dithered-image>
    let containingDiv = document.createElement("div")
    containingDiv.classList.add("imagediv")
    containingDiv.classList.add(cssClass)
    let image = document.createElement("as-dithered-image")
    image.setAttribute("src", "resources/images/" + imageName)
    image.setAttribute("alt", description)
    image.setAttribute("crunch", "pixel")
    image.setAttribute("lightrgba", "rgba(255, 255, 255, 0)")

    containingDiv.appendChild(image)

    parentElement.appendChild(containingDiv)
}

function revealElementAfterDelay(element, delay) {
    element.classList.add("initiallyHidden");
    setTimeout(function () {
        element.classList.remove("initiallyHidden")
        element.classList.add("fadedIn")
    }, delay);
}

let inDialog = false

function continueStory(story, parentElement) {
    // Generate story text - loop through available content
    let firstElement = undefined
    let numParagraphsAdded = 0
    let pageWasEmpty = false

    while (story.canContinue) {

        // generate the next paragraphs
        const paragraphText = story.Continue()
        const tags = story.currentTags

        let elementType = "p"

        tags.forEach((tag) => {
            if (tag === "NEWSECTOR") {
                elementType = "h3"
                removeEverythingFromPage(parentElement)
                firstElement = undefined
                pageWasEmpty = true
            }
            if (tag === "STARTDIALOG") {
                inDialog = true
            }
            if (tag === "ENDDIALOG") {
                inDialog = false
            }
            if (tag.startsWith("SMALLIMAGE")) {
                let matches = tag.match(/SMALLIMAGE (.*)/s)
                console.log(matches)
                insertImage(parentElement, matches[1], "", "smallimage")
                if (firstElement == undefined) {
                    firstElement = parentElement.lastChild
                }
            }
            if (tag.startsWith("IMAGE")) {
                let matches = tag.match(/IMAGE (.*)/s)
                console.log(matches)
                insertImage(parentElement, matches[1], "", "largeimage")
                if (firstElement == undefined) {
                    firstElement = parentElement.lastChild
                }
            }
            if (tag === "EPILOG") {
                let hr = document.createElement("hr")
                hr.classList.add("epilog")
                parentElement.appendChild(hr)
            }
        })

        let paragraphElement;
        if (!inDialog) {
            // should this be just innerText - innerHTML allows for formatting which may be useful
            paragraphElement = document.createElement(elementType)
            paragraphElement.innerHTML = paragraphText
        } else {
            paragraphElement = createDialogElement(paragraphText)
        }
        if (firstElement == undefined) {
            firstElement = paragraphElement
        }
        parentElement.appendChild(paragraphElement)
        revealElementAfterDelay(paragraphElement, numParagraphsAdded * 200);
        numParagraphsAdded++

        if (story.variablesState["inNewSector"] == true) {
            // select a new knot to jump into
            story.variablesState["inNewSector"] = false
            gameState.jumpToNewSector()
        }
    }

    // now generate the choices
    const choiceListTemplate = document.getElementById("choiceList")
    const choiceTemplate = document.getElementById("choice")

    const choiceList = choiceListTemplate.content.cloneNode(true).children[0]
    story.currentChoices.forEach((choice) => {
        const choiceElement = choiceTemplate.content.cloneNode(true).children[0]
        const choiceLink = choiceElement.getElementsByTagName("a")[0]

        choiceLink.innerText = choice.text
        choiceLink.setAttribute("href", "#" + choice.index)
        choiceList.appendChild(choiceElement)

        choiceLink.addEventListener("click", (event) => {
            event.preventDefault()
            choiceClicked(story, event.target)
        })
    })
    parentElement.appendChild(choiceList)
    revealElementAfterDelay(choiceList, numParagraphsAdded * 200)

    if (pageWasEmpty) {
        window.scrollTo(0, 0)
    } else if (firstElement != undefined) {
        firstElement.scrollIntoView({ behavior: "smooth", block: "start", inline: "nearest" })
    }
}

function choiceClicked(story, choiceElement) {
    removeChoicesFromPage(document.getElementById("maintext"))

    let index = choiceElement.getAttribute("href")
    index = index.substring(1)
    const choiceIndex = parseInt(index)

    story.ChooseChoiceIndex(choiceIndex)

    // maybe save state here
    // savePoint = story.state.toJson();

    continueStory(story, document.getElementById("maintext"))
}

function getDirectionInfo(shipSector, direction) {
    let t = gameState.canMove(shipSector, direction)
    return t
}

function displayVariable(id, value) {
    let element = document.getElementById(id)
    if (element) {
        element.textContent = value
    }
}

function initDisplayArtifactVariable(id) {
    displayArtifactVariable(id, gameState.story.variablesState[id])
    gameState.story.ObserveVariable(id, function (s, v) { displayArtifactVariable(id, v); })
}

function displayArtifactVariable(id, value) {
    let element = document.getElementById(id)
    if (element) {
        if (value != 0) {
            element.style.display = "list-item";
            element.textContent = gameState.story.variablesState[id + "_description"]
        } else {
            element.style.display = "none";
            element.textContent = ""
        }
    }
}


/// utility function to create an element in one hit
function createCompleteElement(type, cssClass, text) {
    let result = document.createElement(type)
    result.classList.add(cssClass)
    result.innerText = text
    return result
}

function displayDeaths(parentId, value) {
    // chop up a string the looks like "|value:stardate:details|value:stardate:details|value:stardate:details"
    let lines = value.split("|")
    let deaths = new Array()
    // parse the string
    lines.forEach(l => {
        if (l.length > 0) {
            let sections = l.split(":")
            deaths.push({
                value: parseInt(sections[0]) * -1,
                stardate: parseFloat(sections[1]),
                details: sections[2]
            })
        }
    })
    // sort by stardate (should already be in order, but just in case)
    deaths.sort((left, right) => {
        return left.stardate - right.stardate
    })

    // remove old elements then create elements
    let deathElement = document.getElementById(parentId)
    while (deathElement.firstChild) {
        deathElement.removeChild(deathElement.lastChild)
    }

    if (deaths.length == 0) {
        let e = createCompleteElement("div", "deathReportLine", "No causalities")
        deathElement.appendChild(e)
    } else {
        deaths.forEach(d => {
            deathElement.appendChild(createCompleteElement("span", "date", gameState.formatStardate(d.stardate)))
            deathElement.appendChild(createCompleteElement("span", "value", d.value))
            deathElement.appendChild(createCompleteElement("span", "details", d.details))
        })
    }
}

// we have to start somewhere
async function initSetup() {
    const buttons = document.getElementsByName("displaymode")
    buttons.forEach((e) => {
        e.addEventListener("change", setViewportForButton)
    })

    // setup story
    const storyPath = import.meta.url.replace("gamerunner.js", "storycontent.json")
    let response = await fetch(storyPath)
    let storyContent = await response.json()
    let story = new inkjs.Story(storyContent)

    gameState = new GameState(story, 5, 5)

    displayVariable("fueldisplay", story.variablesState["fuel"]);
    displayVariable("hulldisplay", story.variablesState["hull"]);
    displayVariable("crewdisplay", story.variablesState["crew"]);
    displayVariable("moraledisplay", story.variablesState["morale"]);
    displayVariable("topedodisplay", story.variablesState["torpedoes"]);

    story.ObserveVariable("stardate", function (s, v) { console.log(s, v) })
    story.ObserveVariable("fuel", function (s, v) { displayVariable("fueldisplay", v); })
    story.ObserveVariable("hull", function (s, v) { displayVariable("hulldisplay", v); })
    story.ObserveVariable("crew", function (s, v) { displayVariable("crewdisplay", v); })
    story.ObserveVariable("crewLog", function (s, v) { displayDeaths("crewdeathlog", v) })
    story.ObserveVariable("morale", function (s, v) { displayVariable("moraledisplay", v); })
    story.ObserveVariable("torpedoes", function (s, v) { displayVariable("torpedodisplay", v); })

    initDisplayArtifactVariable("artifact_shield");
    initDisplayArtifactVariable("artifact_weapon");
    initDisplayArtifactVariable("artifact_entertainment");
    initDisplayArtifactVariable("artifact_cloner");
    initDisplayArtifactVariable("artifact_mapper");

    story.BindExternalFunction("getDirectionInfo", getDirectionInfo)
    story.BindExternalFunction("getVisited", gameState.isVisited.bind(gameState))
    story.BindExternalFunction("jumpToNewSector", gameState.jumpToNewSector.bind(gameState))
    story.BindExternalFunction("revealSectorsSurroundingShip", gameState.revealSectorsSurroundingShip.bind(gameState))
    story.BindExternalFunction("findRandomAdjacentSector", gameState.findRandomAdjacentSector.bind(gameState))
    story.BindExternalFunction("getSectorAtOffset", gameState.getSectorAtOffset.bind(gameState))
    story.BindExternalFunction("revealSector", gameState.revealSector.bind(gameState))
    story.BindExternalFunction("isWayBlocked", gameState.isWayBlocked.bind(gameState))
    story.BindExternalFunction("removeMazeWall", gameState.removeMazeWall.bind(gameState))
    story.BindExternalFunction("formatStardate", gameState.formatStardate.bind(gameState))
    story.BindExternalFunction("findSuitableWormholeExit", gameState.findSuitableWormholeExit.bind(gameState))
    story.BindExternalFunction("getBestExitFromSector", gameState.getBestExitFromSector.bind(gameState))
    story.BindExternalFunction("jumpedToPacifica", gameState.jumpedToPacifica.bind(gameState))

    story.BindExternalFunction("initShuffledDeck", Cards.initShuffledDeck)
    story.BindExternalFunction("drawCards", Cards.drawCards)
    story.BindExternalFunction("discardHand", Cards.discardHand)
    story.BindExternalFunction("getHand", Cards.getHand)
    story.BindExternalFunction("handContains", Cards.handContains)

    let parentElement = document.getElementById("maintext")
    story.ChoosePathString("main_title")

    continueStory(story, parentElement)
    starChartDisplay = new StarChart(document.getElementById("starmap"), gameState)
    starChartDisplay.draw()
}


document.addEventListener("DOMContentLoaded", initSetup)