
// random engine from the internet
// used so that the cells get the same random points each time they are generated
function mulberry32(a) {
    return function () {
        var t = a += 0x6D2B79F5;
        t = Math.imul(t ^ t >>> 15, t | 1);
        t ^= t + Math.imul(t ^ t >>> 7, t | 61);
        return ((t ^ t >>> 14) >>> 0) / 4294967296;
    }
}

function getPoints(x, y, up, right, down, left) {
    let rando = mulberry32(x * 517 + y)
    let angle = 2.0 / 3.0 * (2 * Math.PI)
    let radius = 0.30 + ((rando()) * 0.1)
    let radiusRandom = 0.05 + (rando() * 0.05)

    let points = ""
    const divisions = 48
    const passageSize = (2 * Math.PI) / 22.0
    const upAngle = (1.5 * Math.PI)
    const downAngle = (0.5 * Math.PI)
    const leftAngle = Math.PI


    for (let i = 0; i < divisions; ++i) {
        let randomFactor = (rando() - 0.5) * radiusRandom
        let x = (radius + randomFactor) * Math.cos(angle)
        let y = (radius + randomFactor) * Math.sin(angle)

        if (up) {
            let angleDiff = angle - upAngle
            if (Math.abs(angleDiff) < passageSize) {
                y = -0.5
            }
        }
        if (down) {
            let angleDiff = angle - downAngle
            if (Math.abs(angleDiff) < passageSize) {
                y = 0.5
            }
        }
        if (left) {
            let angleDiff = angle - leftAngle
            if (Math.abs(angleDiff) < passageSize) {
                x = -0.5
            }
        }
        if (right) {
            if (angle < passageSize || (angle > (2 * Math.PI - passageSize))) {
                x = +0.5
            }
        }

        points += x.toFixed(3) + "," + y.toFixed(3) + " "
        angle += 2 * Math.PI / divisions
        if (angle > 2 * Math.PI) {
            angle = angle - (2 * Math.PI)
        }
    }
    return points
}


export class StarChart {
    constructor(svgElement, gameState) {
        this.element = svgElement
        this.gameState = gameState
        this.sizeX = this.gameState.getDimensions().x
        this.sizeY = this.gameState.getDimensions().y

        let originalCell = this.element.getElementById("celltemplate")
        let cellContainer = this.element.getElementById("cellcontainer")
        for (let x = 0; x < this.sizeX; ++x) {
            for (let y = 0; y < this.sizeY; ++y) {
                let newCell = originalCell.cloneNode(true)
                newCell.setAttribute("transform", "translate(" + ((x * 20) + 30) + " " + ((y * 20) + 10) + ") scale(20 20)")
                newCell.setAttribute("id", this.getMapCellId(x, y))
                newCell.getElementsByTagName("text")[0].textContent = this.gameState.convertXYToSector({ x: x, y: y })
                cellContainer.appendChild(newCell)
            }
        }
    }

    getMapCellId(x, y) {
        return "mapcell" + x.toFixed(0) + "x" + y.toFixed(0);
    }

    getMapCellElement(x, y) {
        return this.element.getElementById(this.getMapCellId(x, y));
    }

    getMapCellCoords(x, y) {
        return { x: ((x * 20) + 30), y: ((y * 20) + 10) }
    }

    addShipPathElement(parent, id, definition, css, transform) {
        if (this.element.getElementById(id) != null) {
            return; // already exists
        }
        let p = document.createElementNS("http://www.w3.org/2000/svg", "path")
        p.setAttributeNS(null, "d", definition)
        p.id = id
        if (transform != undefined) {
            p.setAttributeNS(null, "transform", transform)
        }

        p.classList.add(css)
        parent.appendChild(p)
    }

    draw() {
        let originalCell = this.element.getElementById("celltemplate")
        for (let x = 0; x < this.sizeX; ++x) {
            for (let y = 0; y < this.sizeY; ++y) {
                let entranceWay = false
                let exitWay = false
                if (x == 0 && (y == this.sizeY - 1)) {
                    entranceWay = true
                }
                if (y == 0 && (x == this.sizeX - 1)) {
                    exitWay = true
                }

                let sector = this.gameState.getSectorAt(x, y)
                let cell = this.getMapCellElement(x, y)
                // cell.getElementsByTagName("text")[1].textContent = sector.knot // debugging
                if (sector.layoutRevealed !== true) {
                    continue;
                }
                let points = getPoints(x, y, (sector.connections.up != undefined), (sector.connections.right != undefined) || exitWay, (sector.connections.down != undefined), (sector.connections.left != undefined) || entranceWay)
                cell.getElementsByTagName("polygon")[0].setAttribute("points", points)
            }
        }

        let shipPath = this.gameState.getShipPath()
        let firstSegment = true

        let shipPathContainer = this.element.getElementById("shippathcontainer")
        let shipOnMap = true

        if (shipPath.length != 0) {
            this.addShipPathElement(shipPathContainer, "entrypath", "M 1 88 c 10,2 22,2, 25,2 m -1 -1 l 1 1 l -1 1", "shipPath")
        } else {
            shipOnMap = false
        }

        for (let i = 0; i < (shipPath.length - 1); ++i) {
            let startSector = shipPath[i]
            let endSector = shipPath[i + 1]
            let start = this.getMapCellCoords(startSector.x, startSector.y)
            let end = this.getMapCellCoords(endSector.x, endSector.y)
            let style = endSector.style
            let dx = endSector.x - startSector.x
            let dy = endSector.y - startSector.y
            let rando = mulberry32(startSector.x * 10 + startSector.y * 23 + endSector.x * 17 + endSector.y * 5 + i)

            if (style == "normal") {
                let angle = ((Math.atan2(dy, dx) * 180) / Math.PI)
                let angleOffset = (rando() * 3)
                angle += angleOffset
                let smallOffset = (rando() - 0.5) * 0.05
                console.log(`${dx} ${dy} ${angle}`)
                let s = `M ${0.2 + smallOffset}, 0.0 l ${0.6 + smallOffset}, 0.0 m -0.05, -0.05 l 0.05, 0.05 l -0.05 0.05`
                this.addShipPathElement(shipPathContainer,
                    `pathpart${i}`,
                    s,
                    "shipPath",
                    `translate(${start.x} ${start.y}) rotate(${angle}) scale(20 20)`)
            } else if (style == "special") {
                // draw a wavey line from start sector to end sector
                const angle = ((Math.atan2(dy, dx) * 180) / Math.PI)
                const length = Math.sqrt(dx * dx + dy * dy) - 0.4
                const segment = length / 4
                const curveAmount = 0.4
                let s = `M 0.2 0.0 q ${segment} ${-curveAmount} ${segment * 2} 0.0 t ${segment * 2} 0.0`
                this.addShipPathElement(shipPathContainer,
                    `pathpart${i}`,
                    s,
                    "shipPathSpecial",
                    `translate(${start.x} ${start.y}) rotate(${angle}) scale(20 20)`)
            } else if (style == "pacifica") {
                shipOnMap = false
                this.addShipPathElement(shipPathContainer, "exitpath", "M 116 10 c 10,-2, 22, -2, 23,-2 m -1 -1 l 1 1 l -1 1", "shipPath")
            }
        }

        // draw the ship marker (if the ship is on the map)
        if (shipOnMap) {
            let shipSector = this.gameState.getCurrentSectorXY()
            let shipMarker = this.element.getElementById("shipmarker")
            shipMarker.setAttribute("transform", "translate(" + ((shipSector.x * 20) + 30) + " " + ((shipSector.y * 20) + 10) + ") scale(20 20)")
            shipMarker.setAttribute("visibility", "visible")
        } else {
            let shipMarker = this.element.getElementById("shipmarker")
            shipMarker.setAttribute("visibility", "hidden")
        }
    }
}
